/************************************************
*	dsalloc - the low level support functions	*
*	for the data structure toolbox.				*
************************************************/

#define IN_SOURCE_DSALLOC

#include <stdio.h>		/* NULL, fprintf		*/
#include <alloc.h>		/* free, malloc			*/
#include <mem.h>		/* memcpy				*/
#include "dsdebug.h"	/* dsChecks, dsStat &	*/
						/* function prototypes	*/

#define  TRUE	  (1==1)
#define  MIN(a,b) (((a)>(b))?(b):(a))

/************************************************
* Structure of an allocated memory node			*
*												*
*	type &										*
*	 size		Description						*
*	-------		----------------------------	*
*	int (2)		size of the block (in bytes)	*
*	int (2)		sequence number of block		*
*	ptr (4)		next allocated block			*
*	ptr (4)		previously allocated block		*
*	? ? ? ?		USER DATA (even # bytes in size)*
*	int (2)		check word (for validation)		*
************************************************/
#define DS_SZPR 6		/* preamble size (# ints*/
typedef union			/** Pointer Definition **/
	{
	int far *ip;		/* Map ptr to block of	*/
	struct tMN far *mnp; /* ints to MemNode ptr	*/
	} dsMNptr;
struct tMN				/**** Block Preamble ****/
	{
	int siz;			/* Size of USER's data	*/
	int seq;			/* Alloc sequence number*/
	dsMNptr nxt;		/* Ptr to next node		*/
	dsMNptr prv;		/* Ptr to previous node */
	};

/************************************************
* Local data for the allocation/free routines	*
************************************************/
dsMNptr dsAlast={(int far *)fNULL};	/*Last node	*/
dsMNptr dsAroot={(int far *)fNULL};	/*First node*/
dsStat _dsErrType=dssOK;	/* Last ERROR found	*/
static int dsAcnt=1;		/* Block number		*/

/************************************************
* dsAcheck - check/set the check word			*
*												*
* INP:	p - address of block to check			*
*		m - operation (set, reset or check)		*
* OUT:	dssBADCHECK if check fails, otherwise	*
*		dssOK									*
************************************************/
#define CheckByte x.ip[x.mnp->siz/2+DS_SZPR]
dsStat near dsAcheck(dsMNptr x, dsChecks m)
	{
	int ck, i;

	/* Compute the XOR remainder of preamble */
	for (ck=i=0; i<DS_SZPR; i++)
		ck ^= x.ip[i];
	ck ^= 0x8421;

	/* What to do with the remainder ... */
	switch (m)
		{
		case dcSET:		/* Make PREAMBLE valid	*/
				CheckByte = ck;
				break;
		case dcRESET:	/* Make PREAMBLE illegal*/
				CheckByte = ck+1;
				break;
		case dcCHECK:	/* Check validity		*/
				if (CheckByte != ck)
					{
					_dsErrType=dssBADCHECK;
					return _dsErrType;
					}
				break;
		default:		/* Unknown request		*/
				_dsErrType=dssBADREQ;
				return _dsErrType;
				break;
		}
	return dssOK;
	}
#undef CheckByte

/************************************************
* dsAdump - dump the list of allocated blocks	*
* of memory to the indicated stream.			*
*												*
* INP:	s - stream to send output to (stderr if	*
*			s == NULL)							*
* OUT:	dsOK if no errors are found in data		*
************************************************/
dsStat far dsAdump(FILE *s)
	{
	int		Corrupt, cnt;
	dsMNptr x;

	if (s==NULL) s=stderr;
	Corrupt = cnt = 0;
	fprintf(s,"node addr   size   seq #  next node"
			"  prev node  Good?\n---------  ------"
			"  -----  ---------  ---------  -----"
			"\n");
	for (x=dsAroot;x.ip!=(int far *)fNULL;
	x=x.mnp->nxt)
		{
		if (dsAcheck(x,dcCHECK)==dssOK)
			Corrupt=' ';
		else
			{
			Corrupt='N';
			if (++cnt > MAXCORR)
				return _dsErrType=dssBADMEM;
			}
		fprintf(s,"%Fp  %6d  %5d  %Fp  %Fp   %c\n",
			x,x.mnp->siz,x.mnp->seq,x.mnp->nxt,
			x.mnp->prv,Corrupt);
		}
	return _dsErrType;
	}

/************************************************
* dsAlloc - internal routine used to actually	*
* allocate new blocks of RAM					*
*												*
* INP:	size - size of the item to get			*
* OUT:	NULL on error, otherwise a pointer to	*
*		a block of RAM the requested size.  Use	*
*		dsLastError to get the error if NULL.	*
************************************************/
static void far * far dsAlloc(size_t size)
	{
	dsMNptr t;

	/* Create the new node */
	size += size&1;			/* Make size even	*/
	if (!(t.ip=(int far *)farmalloc(size+(DS_SZPR
	+1)*2)))
		{
		_dsErrType = dssNOMEM;
		return fNULL;
		}
	t.mnp->nxt.ip = (int far *)fNULL;
	t.mnp->prv.ip = (int far *)fNULL;
	t.mnp->seq = dsAcnt++;
	t.mnp->siz = size;

	/* Link to the Alloc chain */
	if (dsAroot.ip == (int far *)fNULL)
		dsAroot=t;
	else
		{
		t.mnp->prv.mnp = dsAlast.mnp;
		dsAlast.mnp->nxt=t;
		dsAcheck(dsAlast,dcSET);
		}
	dsAlast=t;

	dsAcheck(t,dcSET);		/* Make block valid	*/
	return t.ip+DS_SZPR;	/* return it		*/
	}

/************************************************
* dsCalloc - allocate a chunk of RAM to store	*
* an array of items.							*
*												*
* INP:	n - number of items						*
*		m - size of each item (in bytes)		*
* OUT:	NULL - error occurred (use dsLastError	*
*		to find out what); otherwise, it will	*
*		return the address of the initialized	*
*		block.									*
************************************************/
void far * far dsCalloc(size_t n, size_t m)
	{
	/* Check for errors in Handle or Block list	*/
	if (dsValid(fNULL,"CALLOC",NULL) != dssOK)
		return fNULL;

	return dsAlloc(n*m);
	}

/************************************************
* dsFree - free (release) a block of previously	*
* allocated ram (from dsAlloc).					*
*												*
* INP:	p - block pointer to be released		*
* OUT:	status (see def. of dsStat in datstruc.h*
************************************************/
dsStat far dsFree(void far *p)
	{
	dsMNptr x;

	/* Check for errors in Handle or Block list	*/
	if (dsValid(p,"FREE",NULL) != dssOK)
		return _dsErrType;

	x.ip = ((int far *)p)-DS_SZPR;

	/* Get rid of the node */
	dsAcheck(x,dcRESET);	/* Invalidate block	*/
	if (x.mnp->prv.ip)		/* remove from chain*/
		{
		x.mnp->prv.mnp->nxt=x.mnp->nxt;
		dsAcheck(x.mnp->prv,dcSET);
		}
	if (x.mnp->nxt.ip)
		{
		x.mnp->nxt.mnp->prv=x.mnp->prv;
		dsAcheck(x.mnp->nxt,dcSET);
		}
	if (dsAroot.ip == x.ip)
		dsAroot.ip = x.mnp->nxt.ip;
	if (dsAlast.ip == x.ip)
		dsAlast.ip = x.mnp->prv.ip;
	farfree((void far *)x.ip);	/* Release it	*/
	return dssOK;
	}

/************************************************
* dsLastError - return the LAST ERROR code (not	*
* dssOK) found by any ds??? routine				*
*												*
* OUT:	The last error code found.  If dssOK,	*
*		then no errors have ever been given.	*
************************************************/
dsStat far dsLastError(void)
	{
	return _dsErrType;
	}

/************************************************
* dsMalloc - allocate a chunk of RAM			*
*												*
* INP:	s - size (in bytes) of desired chunk	*
* OUT:	NULL - indicates an error was found,	*
*		use dsLastError to find which one.		*
* NOTE: Each invocation checks the last node	*
*		allocated for errors.					*
************************************************/
void far * far dsMalloc(size_t s)
	{
	/* Check for errors in Handle or Block list	*/
	if (dsValid(fNULL,"MALLOC",NULL)
	!= dssOK)
		return fNULL;

	return dsAlloc(s);
	}

/************************************************
* dsRealloc - reallocate a specified block.		*
*												*
* INP:	p - handle of the BLOCK to be resized.	*
*		s - new size of the block.				*
* OUT:	The address of the BLOCK, or NULL if an	*
*		error is detected.						*
* NOTE:	The output address is ALWAYS different	*
*		from the input one.  This is because:	*
*		(1) It was easier this way, and			*
*		(2) It enhances debugging! (yea, right)	*
************************************************/
void far * far dsRealloc(void far *p, size_t s)
	{
	void far *t;
	dsMNptr x;

	t = dsAlloc(s);				/* Get new block*/
	if (t==fNULL)				/* check for err*/
		return fNULL;
	x.ip = (int far *)p-DS_SZPR;/* Copy data	*/
	memcpy(t,p,(size_t)MIN(s,x.mnp->siz));
	if (dsFree(p)!=dssOK)		/* Kill old blk	*/
		return fNULL;
	else
		return t;
	}

/************************************************
* dsValid - check if the handle passed in is	*
* legal (i.e. in the list of memory blocks),	*
* and check the list for corruption.			*
*												*
* INP:	p - handle of USER BLOCK to check (if	*
*			NULL, then ignore this check)		*
*		m - additional message to print on error*
*		s - stream to output error messages to	*
*			(If NULL, use stderr)				*
* OUT:	the last error code encountered, or 	*
*		dssOK if no error was encountered.		*
* NOTE: Only the FIRST memory block overrun will*
*		be detailed.  The rest will be counted,	*
*		however.								*
* NOTE: Only ONE error may be returned.  The one*
*		with the highest priority will be the	*
*		one returned.  The priorities are:		*
*			1 - Bad block list					*
*			2 - Invalid handle					*
*			3 - Corruption found in list		*
************************************************/
static char NULLmsg[]="";
dsStat far dsValid(void far *p, char *m, FILE *s)
	{
	int		Corrupt,		/* # bad blocks fnd */
			HandleFound;
	dsMNptr x;

	/** Set default values, if necessary **/
	if (s==NULL) s=stderr;
	if (m==NULL) m=NULLmsg;
	HandleFound = (p==fNULL);

	/** Check memory chain & find desired block */
	for (Corrupt=0, x=dsAroot; x.ip!=fNULL;
	x=x.mnp->nxt)
		{
		/* Have we found the handle yet? */
		if ((int far *)p==x.ip+DS_SZPR)
			HandleFound = TRUE;

		/* Is the memory block corrupt? */
		if (dsAcheck(x,dcCHECK) != dssOK)
			if (++Corrupt > MAXCORR)
				{
				fprintf(s,"ERROR: Bad List\7\n");
				return _dsErrType=dssBADMEM;
				}
			else
				fprintf(s,"ERROR: block %FpH cor"
					"rupt.\n\t%s\n\tBlk # %d, si"
					"ze %d.\n\7",x.ip+DS_SZPR,m,
					x.mnp->seq,x.mnp->siz);
		}
	if (Corrupt>1)
		fprintf(s,"\t%d corrupt blocks found.\n",
			Corrupt);

	/** Determine return code & further errors **/
	if (HandleFound)
		if (Corrupt)
			return _dsErrType=dssBADCHECK;
		else
			return dssOK;
	else
		{
		fprintf(s,"ERROR: requested block (%Fp) "
			"not found.\n\7",p);
		if (Corrupt)
			fprintf(s,"\tPerhaps from block corr"
				"uption.\n");
		return _dsErrType=dssBADHANDLE;
		}
	}
